"""For more information about packaging, check from the below.

https://packaging.python.org/tutorials/packaging-projects/
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Omnis",
    version="0.0.10.37",
    author="Omnis Labs Company",
    author_email="mkh48v@snu.ac.kr",
    description="Deep Learning for everyone",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/omnis-labs-company/omnis",
    install_requires=[
        'opencv-python>=3.4.2.17',
        'tensorflow>=1.12.0',
        'keras>=2.2.4',
        'matplotlib>=3.1.0',
        'numpy>=1.14.3',
        'scikit-image>=0.13.1',
        'scipy>=1.3.0',
        'Pillow>=6.0.0',
        'pandas>=0.24.2'
    ],
    classifiers=(
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    packages=setuptools.find_packages(),
)



"""
Mask R-CNN
Base Configurations class.

Copyright (c) 2017 Matterport, Inc.
Licensed under the MIT License (see LICENSE for details)
Written by Waleed Abdulla
"""

import numpy as np


# Base Configuration Class
# Don't use this class directly. Instead, sub-class it and override
# the configurations you need to change.

class Config(object):
    """Base configuration class. For custom configurations, create a
    sub-class that inherits from this one and override properties
    that need to be changed.
    """
    # Name the configurations. For example, 'COCO', 'Experiment 3', ...etc.
    # Useful if your code needs to do things differently depending on which
    # experiment is running.
    NAME = None  # Override in sub-classes

    # NUMBER OF GPUs to use. When using only a CPU, this needs to be set to 1.
    GPU_COUNT = 1

    # Number of images to train with on each GPU. A 12GB GPU can typically
    # handle 2 images of 1024x1024px.
    # Adjust based on your GPU memory and image sizes. Use the highest
    # number that your GPU can handle for best performance.
    IMAGES_PER_GPU = 2

    # Number of training steps per epoch
    # This doesn't need to match the size of the training set. Tensorboard
    # updates are saved at the end of each epoch, so setting this to a
    # smaller number means getting more frequent TensorBoard updates.
    # Validation stats are also calculated at each epoch end and they
    # might take a while, so don't set this too small to avoid spending
    # a lot of time on validation stats.
    STEPS_PER_EPOCH = 1000

    # Number of validation steps to run at the end of every training epoch.
    # A bigger number improves accuracy of validation stats, but slows
    # down the training.
    VALIDATION_STEPS = 50

    # Backbone network architecture
    # Supported values are: resnet50, resnet101.
    # You can also provide a callable that should have the signature
    # of model.resnet_graph. If you do so, you need to supply a callable
    # to COMPUTE_BACKBONE_SHAPE as well
    BACKBONE = "resnet101"

    # Only useful if you supply a callable to BACKBONE. Should compute
    # the shape of each layer of the FPN Pyramid.
    # See model.compute_backbone_shapes
    COMPUTE_BACKBONE_SHAPE = None

    # The strides of each layer of the FPN Pyramid. These values
    # are based on a Resnet101 backbone.
    BACKBONE_STRIDES = [4, 8, 16, 32, 64]

    # Size of the fully-connected layers in the classification graph
    FPN_CLASSIF_FC_LAYERS_SIZE = 1024

    # Size of the top-down layers used to build the feature pyramid
    TOP_DOWN_PYRAMID_SIZE = 256

    # Number of classification classes (including background)
    NUM_CLASSES = 1  # Override in sub-classes

    # Length of square anchor side in pixels
    RPN_ANCHOR_SCALES = (32, 64, 128, 256, 512)

    # Ratios of anchors at each cell (width/height)
    # A value of 1 represents a square anchor, and 0.5 is a wide anchor
    RPN_ANCHOR_RATIOS = [0.5, 1, 2]

    # Anchor stride
    # If 1 then anchors are created for each cell in the backbone feature map.
    # If 2, then anchors are created for every other cell, and so on.
    RPN_ANCHOR_STRIDE = 1

    # Non-max suppression threshold to filter RPN proposals.
    # You can increase this during training to generate more propsals.
    RPN_NMS_THRESHOLD = 0.7

    # How many anchors per image to use for RPN training
    RPN_TRAIN_ANCHORS_PER_IMAGE = 256

    # ROIs kept after tf.nn.top_k and before non-maximum suppression
    PRE_NMS_LIMIT = 6000

    # ROIs kept after non-maximum suppression (training and inference)
    POST_NMS_ROIS_TRAINING = 2000
    POST_NMS_ROIS_INFERENCE = 1000

    # If enabled, resizes instance masks to a smaller size to reduce
    # memory load. Recommended when using high-resolution images.
    USE_MINI_MASK = True
    MINI_MASK_SHAPE = (56, 56)  # (height, width) of the mini-mask

    # Input image resizing
    # Generally, use the "square" resizing mode for training and predicting
    # and it should work well in most cases. In this mode, images are scaled
    # up such that the small side is = IMAGE_MIN_DIM, but ensuring that the
    # scaling doesn't make the long side > IMAGE_MAX_DIM. Then the image is
    # padded with zeros to make it a square so multiple images can be put
    # in one batch.
    # Available resizing modes:
    # none:   No resizing or padding. Return the image unchanged.
    # square: Resize and pad with zeros to get a square image
    #         of size [max_dim, max_dim].
    # pad64:  Pads width and height with zeros to make them multiples of 64.
    #         If IMAGE_MIN_DIM or IMAGE_MIN_SCALE are not None, then it scales
    #         up before padding. IMAGE_MAX_DIM is ignored in this mode.
    #         The multiple of 64 is needed to ensure smooth scaling of feature
    #         maps up and down the 6 levels of the FPN pyramid (2**6=64).
    # crop:   Picks random crops from the image. First, scales the image based
    #         on IMAGE_MIN_DIM and IMAGE_MIN_SCALE, then picks a random crop of
    #         size IMAGE_MIN_DIM x IMAGE_MIN_DIM. Can be used in training only.
    #         IMAGE_MAX_DIM is not used in this mode.
    IMAGE_RESIZE_MODE = "square"
    IMAGE_MIN_DIM = 800
    IMAGE_MAX_DIM = 1024
    # Minimum scaling ratio. Checked after MIN_IMAGE_DIM and can force further
    # up scaling. For example, if set to 2 then images are scaled up to double
    # the width and height, or more, even if MIN_IMAGE_DIM doesn't require it.
    # However, in 'square' mode, it can be overruled by IMAGE_MAX_DIM.
    IMAGE_MIN_SCALE = 0
    # Number of color channels per image. RGB = 3, greyscale = 1, RGB-D = 4
    # Changing this requires other changes in the code. See the WIKI for more
    # details: https://github.com/matterport/Mask_RCNN/wiki
    IMAGE_CHANNEL_COUNT = 3

    # Image mean (RGB)
    MEAN_PIXEL = np.array([123.7, 116.8, 103.9])

    # Number of ROIs per image to feed to classifier/mask heads
    # The Mask RCNN paper uses 512 but often the RPN doesn't generate
    # enough positive proposals to fill this and keep a positive:negative
    # ratio of 1:3. You can increase the number of proposals by adjusting
    # the RPN NMS threshold.
    TRAIN_ROIS_PER_IMAGE = 200

    # Percent of positive ROIs used to train classifier/mask heads
    ROI_POSITIVE_RATIO = 0.33

    # Pooled ROIs
    POOL_SIZE = 7
    MASK_POOL_SIZE = 14

    # Shape of output mask
    # To change this you also need to change the neural network mask branch
    MASK_SHAPE = [28, 28]

    # Maximum number of ground truth instances to use in one image
    MAX_GT_INSTANCES = 100

    # Bounding box refinement standard deviation for RPN and final detections.
    RPN_BBOX_STD_DEV = np.array([0.1, 0.1, 0.2, 0.2])
    BBOX_STD_DEV = np.array([0.1, 0.1, 0.2, 0.2])

    # Max number of final detections
    DETECTION_MAX_INSTANCES = 100

    # Minimum probability value to accept a detected instance
    # ROIs below this threshold are skipped
    DETECTION_MIN_CONFIDENCE = 0.7

    # Non-maximum suppression threshold for detection
    DETECTION_NMS_THRESHOLD = 0.3

    # Learning rate and momentum
    # The Mask RCNN paper uses lr=0.02, but on TensorFlow it causes
    # weights to explode. Likely due to differences in optimizer
    # implementation.
    LEARNING_RATE = 0.001
    LEARNING_MOMENTUM = 0.9

    # Weight decay regularization
    WEIGHT_DECAY = 0.0001

    # Loss weights for more precise optimization.
    # Can be used for R-CNN training setup.
    LOSS_WEIGHTS = {
        "rpn_class_loss": 1.,
        "rpn_bbox_loss": 1.,
        "mrcnn_class_loss": 1.,
        "mrcnn_bbox_loss": 1.,
        "mrcnn_mask_loss": 1.
    }

    # Use RPN ROIs or externally generated ROIs for training
    # Keep this True for most situations. Set to False if you want to train
    # the head branches on ROI generated by code rather than the ROIs from
    # the RPN. For example, to debug the classifier head without having to
    # train the RPN.
    USE_RPN_ROIS = True

    # Train or freeze batch normalization layers
    #     None: Train BN layers. This is the normal mode
    #     False: Freeze BN layers. Good when using a small batch size
    #     True: (don't use). Set layer in training mode even when predicting
    TRAIN_BN = False  # Defaulting to False since batch size is often small

    # Gradient norm clipping
    GRADIENT_CLIP_NORM = 5.0

    def __init__(self):
        """Set values of computed attributes."""
        # Effective batch size
        self.BATCH_SIZE = self.IMAGES_PER_GPU * self.GPU_COUNT

        # Input image size
        if self.IMAGE_RESIZE_MODE == "crop":
            self.IMAGE_SHAPE = np.array([self.IMAGE_MIN_DIM, self.IMAGE_MIN_DIM,
                                         self.IMAGE_CHANNEL_COUNT])
        else:
            self.IMAGE_SHAPE = np.array([self.IMAGE_MAX_DIM, self.IMAGE_MAX_DIM,
                                         self.IMAGE_CHANNEL_COUNT])

        # Image meta data length
        # See compose_image_meta() for details
        self.IMAGE_META_SIZE = 1 + 3 + 3 + 4 + 1 + self.NUM_CLASSES

    def display(self):
        """Display Configuration values."""
        print("\nConfigurations:")
        for a in dir(self):
            if not a.startswith("__") and not callable(getattr(self, a)):
                print("{:30} {}".format(a, getattr(self, a)))
        print("\n")



import json
import os
from collections import OrderedDict

import cv2
import numpy as np
from keras.optimizers import Adam

from ...application import Application
from ....lib.config_lib import DeblurConfig, DeepblockLogConfig
from ....lib.general_lib import get_data_path_type
from ....lib.image_lib import (
    get_image_names_in_directory, merge_image, save_image, split_one_image,
)
from ....lib.loss_lib import perceptual_loss, wasserstein_loss
from ....lib.model_lib import (
    discriminator_model, generator_containing_discriminator_multiple_outputs,
    generator_model,
)


class Deblur_GAN(Application):
    def __init__(self, g_model_path=None, d_model_path=None):
        super().__init__()

        if not isinstance(g_model_path, type(None)):
            if not isinstance(d_model_path, type(None)):
                self.create_model()
                self.load(g_model_path=g_model_path, d_model_path=d_model_path)
            else:
                self.create_model()
                self.load_g_model(model_path=g_model_path)

    def train(self,
              data_path,
              epochs=1,
              batch_size=32,
              learning_rate=1E-4,
              metrics='accuracy'):
        if not isinstance(metrics, list):
            metrics = [metrics]

        blur_data_path = os.path.join(data_path, 'blur')
        sharp_data_path = os.path.join(data_path, 'sharp')

        blur_array = self.load_image_array(blur_data_path)
        sharp_array = self.load_image_array(sharp_data_path)

        self.create_and_compile_model(learning_rate=learning_rate)

        self._fit_gan(blur_array=blur_array,
                      sharp_array=sharp_array,
                      batch_size=batch_size,
                      epochs=epochs)

    def predict(self, data_path, batch_size=1,
                result_path=DeblurConfig.RESULT_FOLDER):
        image_data = self.load_predict_images_data(data_path)
        image_array = image_data['images']
        image_paths_list = image_data['images_paths']
        image_area_index = image_data['images_area_index']

        generated_images = self.g_model.predict(
            x=image_array, batch_size=batch_size)
        generated_images_array = np.array(
            [self.deprocess_image(img) for img in generated_images])

        image_list = []
        for i in range(generated_images.shape[0]):
            img = generated_images_array[i, :, :, :]
            image_list.append(img)

        now_image_path = image_paths_list[0]
        merge_image_list = []
        merge_image_area_list = []
        for i, image in enumerate(image_paths_list):
            if image != now_image_path:
                merged_image = merge_image(
                    merge_image_list, merge_image_area_list)
                self.save_merged_image(
                    merged_image, result_path, now_image_path)
                now_image_path = image
                merge_image_list = []
                merge_image_area_list = []
            merge_image_list.append(image_list[i])
            merge_image_area_list.append(image_area_index[i])

            if i == len(image_paths_list) - 1:
                merged_image = merge_image(
                    merge_image_list, merge_image_area_list)
                self.save_merged_image(
                    merged_image, result_path, now_image_path)

    def save(self, g_model_path, d_model_path):
        try:
            self.save_g_model(g_model_path)
            self.save_d_model(d_model_path)
        except Exception as e:
            raise e

    def load(self, g_model_path, d_model_path):
        try:
            self.load_g_model(g_model_path)
            self.load_d_model(d_model_path)
        except Exception as e:
            raise e

    def _fit_gan(self,
                 blur_array,
                 sharp_array,
                 batch_size,
                 epochs,
                 critic_updates=5):
        output_true_batch = np.ones((batch_size, 1))
        output_false_batch = -np.ones((batch_size, 1))

        for epoch in range(0, epochs):
            permutated_indexes = np.random.permutation(
                blur_array.shape[0])

            d_losses = []
            d_on_g_losses = []
            for index in range(int(blur_array.shape[0] / batch_size)):
                batch_indexes = permutated_indexes[
                    index * batch_size
                    :(index + 1) * batch_size
                ]
                image_blur_batch = blur_array[batch_indexes]
                image_full_batch = sharp_array[batch_indexes]

                generated_images = self.g_model.predict(
                    x=image_blur_batch, batch_size=batch_size)

                for _ in range(critic_updates):
                    d_loss_real = self.d_model.train_on_batch(
                        image_full_batch, output_true_batch)
                    d_loss_fake = self.d_model.train_on_batch(
                        generated_images, output_false_batch)
                    d_loss = 0.5 * np.add(d_loss_fake, d_loss_real)
                    d_losses.append(d_loss)

                self.d_model.trainable = False

                d_on_g_loss = self.d_on_g.train_on_batch(
                    image_blur_batch, [image_full_batch, output_true_batch])
                d_on_g_losses.append(d_on_g_loss)

                self.d_model.trainable = True

            if self.deepblock_log:
                log_json = OrderedDict()
                log_json[DeepblockLogConfig.LOG_TYPE] = "epochend"
                log_json['totalEpoch'] = epochs
                log_json['nowEpoch'] = epoch
                log_json['d_loss'] = str(np.mean(d_losses))
                log_json['d_on_g_loss'] = str(np.mean(d_on_g_losses))
                print(json.dumps(log_json, ensure_ascii=False))

    def create_and_compile_model(self,
                                 learning_rate):
        self.create_model()

        d_opt = Adam(lr=learning_rate, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
        d_on_g_opt = Adam(lr=learning_rate, beta_1=0.9, beta_2=0.999,
                          epsilon=1e-08)

        self.d_model.trainable = True
        self.d_model.compile(optimizer=d_opt,
                             loss=wasserstein_loss)
        self.d_model.trainable = False

        self.d_on_g.compile(d_on_g_opt,
                            loss=[perceptual_loss, wasserstein_loss],
                            loss_weights=[100, 1])
        self.d_model.trainable = False

    def create_model(self):
        self.g_model = self.create_g_model()
        self.d_model = self.create_d_model()
        self.d_on_g = generator_containing_discriminator_multiple_outputs(
            self.g_model, self.d_model)

    def create_g_model(self):
        g_model = generator_model()
        return g_model

    def create_d_model(self):
        d_model = discriminator_model()
        return d_model

    def load_g_model(self, model_path):
        self.g_model.load_weights(model_path)

    def load_d_model(self, model_path):
        self.d_model.load_weights(model_path)

    def save_g_model(self, model_path):
        self.g_model.save_weights(model_path)

    def save_d_model(self, model_path):
        self.d_model.save_weights(model_path)

    def save_merged_image(self, image, result_path, image_path):
        if not os.path.exists(result_path):
            os.makedirs(result_path)

        save_image(image, os.path.join(result_path, image_path))
        if self.deepblock_log:
            log_json = OrderedDict()
            log_json[DeepblockLogConfig.LOG_TYPE] = "fileGenerated"
            log_json['filename'] = image_path
            print(json.dumps(log_json, ensure_ascii=False))

    def load_image_array(self, data_path):
        data_path_type = get_data_path_type(data_path)

        if data_path_type == "image":
            all_image_paths = [data_path]
        elif data_path_type == "directory":
            all_image_paths = get_image_names_in_directory(data_path)
        else:
            raise ValueError("path should be image file or directory")

        images_all = []
        for path_image in all_image_paths:
            img = cv2.imread(path_image, cv2.IMREAD_COLOR)
            images_all.append(self.preprocess_image(img))

        return np.array(images_all)

    def load_predict_images_data(self, path):
        if os.path.isfile(path):
            all_image_paths = [path]
        elif os.path.isdir(path):
            all_image_paths = get_image_names_in_directory(path)
        else:
            raise ValueError("path should be file or directory")

        images_all = []
        images_all_paths = []
        images_all_area_index = []
        for path_image in all_image_paths:
            img = cv2.imread(path_image, cv2.IMREAD_COLOR)
            split_images, area_index = split_one_image(
                img, DeblurConfig.CELL_SHAPE[0], DeblurConfig.CELL_SHAPE[1])
            for i, image in enumerate(split_images):
                images_all.append(self.preprocess_image(image))
                images_all_paths.append(path_image)
                images_all_area_index.append(area_index[i])

        return {
            'images': np.array(images_all),
            'images_paths': images_all_paths,
            'images_area_index': images_all_area_index
        }

    def preprocess_image(self, img):
        resized_img = cv2.resize(img, DeblurConfig.CELL_SHAPE[:2])
        processed_img = (resized_img - 127.5) / 127.5
        return processed_img

    def deprocess_image(self, img):
        img = img * 127.5 + 127.5
        return img.astype('uint8')
